(** The core of the BSMLlib library*)

type 'a par
(** Abstract type for parallel vector of size p. In the following comments
we will note <v_0,...,v_(p-1)> the parallel vector whose
value at processor 0 is v_O, value at processor 1 is v_1, {i etc.} *)

val bsp_p: unit -> int
(** [bsp_p()] returns the number {i p} of processes of the parallel machine. *)

val bsp_g: unit -> float
(** [bsp_g()] returns the BSP parameter {i g} of the parallel machine. *)

val bsp_l: unit -> float
(** [bsp_l()] returns the BSP parameter {i l} of the parallel machine. *)

val mkpar: (int -> 'a) -> 'a par
(** [mkpar f] evaluates to the parallel vector <v{_0},...,v{_p-1}>
where v{_i}=[(f i)]. *)

val apply: ('a -> 'b) par -> 'a par -> 'b par
(** [apply vf vv] where [vf] is a parallel vector of functions 
<f{_0},...,f{_p-1}> and [vv] a parallel vector of values <v{_0},...,v{_p-1}>.
evaluates to <w{_0},...,w{_p-1}> where w{_i} is the result of 
the evaluation of [(f{_i} v{_i})]. This operation is called pointwise 
parallel application. *) 

val put: (int -> 'a option) par -> (int -> 'a option) par 
(** Consider the expression [put(mkpar(fun i->fs]{_i}[))] (1).  To send a
value {i v} from process {i j} to process {i i}, the function {i
fs}{_j} at process {i j} must be such as ({i fs}{_j}{i i}) evaluates
to [Some] {i v}.  To send no value from process {i j} to process {i
i}, ({i fs}{_j}{i i}) must evaluate to [None].  Expression (1)
evaluates to a parallel vector containing a function {i fd}{_i} of
delivered messages on every process.  At process {i i}, ({i fd}{_i}{i
j}) evaluates to [None] if process {i j} sent no message to process {i
i} or evaluates to [Some] {i v} if process {i j} sent the value {i v}
to the process {i i}. *)

exception At_failure of string
val at: 'a par -> int -> 'a
(** NO MORE TRUE: [at] must be only used in a [if then else] expression. 
The expressions [if at] <b{_0},...,b{_p-1}> [then] E{_1} [else] E{_2} 
will be evaluated to [E_1] if b{_pid} is [true], 
to [E_2] if b{_pid} is [false]. 
E{_1} and E{_2} must be expressions of global type.
[At_failure] is raised if one tries to access a pid not in the
interval 0,...,{i p}-1 where {i p} is the number of processors of the
machine. *)

val unsafe_proj: 'a par -> 'a 
(** [safe_proj] <v,...,v> = v, otherwise unpredictable and unspeficied
behaviour if the argument has not the same value at each process *)

val initialize: unit -> unit 
(**  [initialize()] initiates a parallel computation.  When used with
the sequential version of the module Bsmllib, it reads the first line
of [$HOME/.bsmllibrc] in order to obtain the values for [bsp_p()], 
[bsp_g()] and [bsp_l()]. When used with the parallel version
of the module Bsmllib, it automatically determines the number
of processors [bsp_p ()], then reads the appropriated line of
[$HOME/.bsmllibrc] in order to obtain the values for [bsp_g]
 and [bsp_l].  A program must have one and only one call to [initialize()]. *)

val abort: int -> string -> unit
(** [(abort error msg)] prints the message [msg] and 
aborts the computation with exit code [error]. *)

exception Timer_failure of string
val start_timing: unit->unit
val stop_timing: unit->unit
val get_cost: unit -> float par
(** [start_timing()] starts the timing. [stop_timing()] stops it. [get_cost()] returns 
a parallel vector which contains for each processor the time elapsed between 
the call of [start_timing] and [stop_timing]. The exception [Timer_failure] is 
raised if the call to one of those functions is meaningless (for eg a call to
[stop_timing] if [start_timing] have not been called before. *)


