
module type TYPE =
 sig
  type 'a t

  val init : int -> (int -> 'a) -> 'a t
  val length : 'a t -> int
  val empty : 'a t
  val map : ('a -> 'b) -> 'a t -> 'b t
  val zip : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
  val zip2 : ('a -> 'b -> 'c -> 'd) -> 'a t -> 'b t -> 'c t -> 'd t
  val to_array : 'a t -> 'a array
  val to_list : 'a t -> 'a list
  val shift_left : 'a -> 'a t -> 'a t
  val shift_right : 'a -> 'a t -> 'a t
  val concat : 'a t list -> 'a t
  val get_last : 'a t -> 'a
  val get_first : 'a t -> 'a
  val sub : 'a t -> int -> int -> 'a t
 end

module Make = functor (T : TYPE) -> functor (Bsml : Bsmlsig.BSML) -> struct
 module Base = Bsmlbase.Make(Bsml)
 module Comm = Bsmlcomm.Make(Bsml)

 let make f size =
  let k = Bsml.mkpar (fun i -> size/Bsml.bsp_p + (if i < (Tools.natmod size Bsml.bsp_p) then 1 else 0)) in
  let k' = Comm.scan_logp (+) k in
  let k'' = Base.parfun2 (-) k' k in
   Base.parfun2 (fun x y -> T.init x (fun z -> f (z + y))) k k''

 let extract a =
  let k = Comm.total_exchange a in
  let k' = Base.parfun T.concat k in
   Bsml.proj k' 0

 let par a =
  let l = T.length a in
  let k = Bsml.mkpar (fun i -> l/Bsml.bsp_p + (if i < (Tools.natmod l Bsml.bsp_p) then 1 else 0)) in
  let k' = Comm.scan_logp (+) k in
  let k'' = Base.parfun2 (-) k' k in
   Base.parfun2 (fun x y -> T.sub a y x) k k''

 let length v =
  let k = Base.parfun T.length v in
  let k' = Comm.scan_logp (+) k in
   Bsml.proj k' (Bsml.bsp_p-1)

 let map f = Base.parfun (T.map f)
 let zip f = Base.parfun2 (T.zip f)

 (* PRIVATE *)
 (* 'a T par -> int T par *)
 let par_indexes v =
  let k = Base.parfun T.length v in
  let k' = Comm.scan_logp (+) k in
  let k'' = Base.parfun2 (-) k' k in
   Base.parfun2 (fun x y -> T.init x (fun z -> z + y)) k k''

 let map_index f v = Base.parfun2 (T.zip f) (par_indexes v) v
 let zip_index f v = Base.parfun3 (T.zip2 f) (par_indexes v) v

 let to_array a = Base.parfun (T.to_array) a
 let to_list a = Base.parfun (T.to_list) a

 let shift_left k v =
  let a = Base.parfun T.get_first v in
  let a' = Base.parfun2 (fun proc i -> if proc = 0 then k else i) Base.this a in
  let a'' = Comm.shift_left a' in
   Base.parfun2 T.shift_left a'' v

 let shift_right k v =
  let a = Base.parfun T.get_last v in
  let a' = Comm.shift_right a in
  let a'' = Base.parfun2 (fun proc i -> if proc = 0 then k else i) Base.this a' in
   Base.parfun2 T.shift_right a'' v

end

module A =
 struct
  type 'a t = 'a array

  let init = Array.init
  let length = Array.length
  let empty = [||]
  let map = Array.map
  let zip f a' b' = Array.init (Array.length a') (fun i -> f a'.(i) b'.(i))
  let zip2 f a' b' c' = Array.init (Array.length a') (fun i -> f a'.(i) b'.(i) c'.(i))
  let to_array = Tools.id
  let to_list = Array.to_list
  let shift_left v a = Array.init (Array.length a) (fun i -> if i = (Array.length a)-1 then v else a.(i+1))
  let shift_right v a = Array.init (Array.length a) (fun i -> if i = 0 then v else a.(i-1))
  let concat = Array.concat
  let get_last v = v.(Array.length v -1)
  let get_first v = v.(0)
  let sub = Array.sub
 end

module L =
 struct
  type 'a t = 'a list

  let length = List.length
  let empty = []
  let map f l =
   let rec aux r = function
    | [] -> r
    | h::t -> aux ((f h)::r) t
   in List.rev (aux [] l)
  let init i f = map f (Tools.from_to 0 (i-1))
  let zip f l l' =
   let rec aux r = function
    | [], [] -> r
    | h::t, h'::t' -> aux ((f h h')::r) (t, t')
     | _ -> raise (Invalid_argument "Skeleton.List.zip")
   in List.rev (aux [] (l, l'))
  let zip2 f a b c =
   let rec aux r = function
     | [],[],[] -> r
     | h1::t1, h2::t2, h3::t3 -> aux ((f h1 h2 h3)::r) (t1, t2, t3)
     | _ -> raise (Invalid_argument "Skeleton.List.zip_index")
   in List.rev (aux [] (a, b, c))
  let to_array = Array.of_list
  let to_list = Tools.id
  let append l l' = List.rev_append (List.rev l) l'
  let shift_left v = function [] -> [v] | l -> append (List.tl l) [v]
  let shift_right v = function [] -> [v] | l -> v :: List.rev (List.tl (List.rev l))
  let concat l = 
   let rec aux r = function
    | [] -> r
    | h::t -> aux (List.rev_append h r) t
   in List.rev (aux [] l)
  let get_last v = List.hd (List.rev v)
  let get_first = List.hd
  let sub v i j =
   let rec aux k l = function
    | [] -> l
    | h::t -> if (k >= i) && (k <= i+j-1) then aux (k+1) (l @ [h]) t
              else if (k > i+j-1) then l
              else aux (k+1) l t
   in aux 0 [] v
 end

module MakeArray = Make(A)
module MakeList = Make(L)

